@@ -1,5 +1,5 @@
 {-# LANGUAGE OverloadedStrings, FlexibleInstances #-}
-{-# LANGUAGE IncoherentInstances, UndecidableInstances #-}
+{-# LANGUAGE LambdaCase #-}
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 
 module Documentation.Haddock.ParserSpec (main, spec) where
@@ -23,12 +23,15 @@ instance IsString (Doc String) where
 instance IsString a => IsString (Maybe a) where
   fromString = Just . fromString
 
-parseParas :: String -> Doc String
-parseParas = Parse.toRegular . Parse.parseParas
+parseParas :: String -> MetaDoc () String
+parseParas = overDoc Parse.toRegular . Parse.parseParas
 
 parseString :: String -> Doc String
 parseString = Parse.toRegular . Parse.parseString
 
+hyperlink :: String -> Maybe String -> Doc String
+hyperlink url = DocHyperlink . Hyperlink url
+
 main :: IO ()
 main = hspec spec
 
@@ -53,8 +56,10 @@ spec = do
       it "accepts hexadecimal character references" $ do
         "&#x65;" `shouldParseTo` "e"
 
-      it "allows to backslash-escape characters" $ do
-        property $ \x -> ['\\', x] `shouldParseTo` DocString [x]
+      it "allows to backslash-escape characters except \\r" $ do
+        property $ \case
+          '\r' -> "\\\r" `shouldParseTo` DocString "\\"
+          x -> ['\\', x] `shouldParseTo` DocString [x]
 
       context "when parsing strings contaning numeric character references" $ do
         it "will implicitly convert digits to characters" $ do
@@ -83,10 +88,13 @@ spec = do
         " don't use apostrophe's in the wrong place's" `shouldParseTo`
           "don't use apostrophe's in the wrong place's"
 
-    context "when parsing URLs" $ do
-      let hyperlink :: String -> Maybe String -> Doc String
-          hyperlink url = DocHyperlink . Hyperlink url
+      it "doesn't parse empty identifiers" $ do
+        "``" `shouldParseTo` "``"
 
+      it "can parse infix identifiers" $ do
+        "``infix``" `shouldParseTo` "`" <> DocIdentifier "infix" <> "`"
+
+    context "when parsing URLs" $ do
       it "parses a URL" $ do
         "<http://example.com/>" `shouldParseTo` hyperlink "http://example.com/" Nothing
 
@@ -115,6 +123,45 @@ spec = do
       it "doesn't allow for multi-line link tags" $ do
         "<ba\nz aar>" `shouldParseTo` "<ba\nz aar>"
 
+      context "when parsing markdown links" $ do
+        it "parses a simple link" $ do
+          "[some label](url)" `shouldParseTo`
+            hyperlink "url" "some label"
+
+        it "allows whitespace between label and URL" $ do
+          "[some label] \t (url)" `shouldParseTo`
+            hyperlink "url" "some label"
+
+        it "allows newlines in label" $ do
+          "[some\n\nlabel](url)" `shouldParseTo`
+            hyperlink "url" "some\n\nlabel"
+
+        it "allows escaping in label" $ do
+          "[some\\] label](url)" `shouldParseTo`
+            hyperlink "url" "some] label"
+
+        it "strips leading and trailing whitespace from label" $ do
+          "[  some label  ](url)" `shouldParseTo`
+            hyperlink "url" "some label"
+
+        it "rejects whitespace in URL" $ do
+          "[some label]( url)" `shouldParseTo`
+            "[some label]( url)"
+
+        context "when URL is on a separate line" $ do
+          it "allows URL to be on a separate line" $ do
+            "[some label]\n(url)" `shouldParseTo`
+              hyperlink "url" "some label"
+
+          it "allows leading whitespace" $ do
+            "[some label]\n  \t (url)" `shouldParseTo`
+              hyperlink "url" "some label"
+
+          it "rejects additional newlines" $ do
+            "[some label]\n\n(url)" `shouldParseTo`
+              "[some label]\n\n(url)"
+
+
       context "when autolinking URLs" $ do
         it "autolinks HTTP URLs" $ do
           "http://example.com/" `shouldParseTo` hyperlink "http://example.com/" Nothing
@@ -145,24 +192,22 @@ spec = do
           "foo https://example.com/example bar" `shouldParseTo`
             "foo " <> hyperlink "https://example.com/example" Nothing <> " bar"
 
-    context "when parsing pictures" $ do
-      let picture :: String -> Maybe String -> Doc String
-          picture uri = DocPic . Picture uri
+    context "when parsing images" $ do
+      let image :: String -> Maybe String -> Doc String
+          image uri = DocPic . Picture uri
 
-      it "parses a simple picture" $ do
-        "<<baz>>" `shouldParseTo` picture "baz" Nothing
+      it "accepts markdown syntax for images" $ do
+        "![label](url)" `shouldParseTo` image "url" "label"
 
-      it "parses a picture with a title" $ do
-        "<<b a z>>" `shouldParseTo` picture "b" (Just "a z")
+      it "accepts Unicode" $ do
+        "![灼眼のシャナ](url)" `shouldParseTo` image "url" "灼眼のシャナ"
 
-      it "parses a picture with unicode" $ do
-        "<<灼眼のシャナ>>" `shouldParseTo` picture "灼眼のシャナ" Nothing
+      it "supports deprecated picture syntax" $ do
+        "<<baz>>" `shouldParseTo` image "baz" Nothing
 
-      it "allows for escaping of the closing tags" $ do
-        "<<ba\\>>z>>" `shouldParseTo` picture "ba>>z" Nothing
+      it "supports title for deprecated picture syntax" $ do
+        "<<b a z>>" `shouldParseTo` image "b" "a z"
 
-      it "doesn't allow for multi-line picture tags" $ do
-        "<<ba\nz aar>>" `shouldParseTo` "<<ba\nz aar>>"
 
     context "when parsing anchors" $ do
       it "parses a single word anchor" $ do
@@ -316,12 +361,39 @@ spec = do
   describe "parseParas" $ do
     let infix 1 `shouldParseTo`
         shouldParseTo :: String -> Doc String -> Expectation
-        shouldParseTo input ast = parseParas input `shouldBe` ast
+        shouldParseTo input ast = _doc (parseParas input) `shouldBe` ast
 
     it "is total" $ do
       property $ \xs ->
         (length . show . parseParas) xs `shouldSatisfy` (> 0)
 
+    context "when parsing @since" $ do
+      it "adds specified version to the result" $ do
+        parseParas "@since 0.5.0" `shouldBe`
+          MetaDoc { _meta = Meta { _version = Just [0,5,0] }
+                  , _doc = DocEmpty }
+
+      it "ignores trailing whitespace" $ do
+        parseParas "@since 0.5.0 \t " `shouldBe`
+          MetaDoc { _meta = Meta { _version = Just [0,5,0] }
+                  , _doc = DocEmpty }
+
+      it "does not allow trailing input" $ do
+        parseParas "@since 0.5.0 foo" `shouldBe`
+          MetaDoc { _meta = Meta { _version = Nothing }
+                  , _doc = DocParagraph "@since 0.5.0 foo" }
+
+
+      context "when given multiple times" $ do
+        it "gives last occurrence precedence" $ do
+          (parseParas . unlines) [
+              "@since 0.5.0"
+            , "@since 0.6.0"
+            , "@since 0.7.0"
+            ] `shouldBe` MetaDoc { _meta = Meta { _version = Just [0,7,0] }
+                                 , _doc = DocEmpty }
+
+
     context "when parsing text paragraphs" $ do
       let filterSpecial = filter (`notElem` (".(=#-[*`\v\f\n\t\r\\\"'_/@<> " :: String))
 
@@ -349,6 +421,28 @@ spec = do
         it "turns it into a code block" $ do
           "@foo@" `shouldParseTo` DocCodeBlock "foo"
 
+      context "when a paragraph starts with a markdown link" $ do
+        it "correctly parses it as a text paragraph (not a definition list)" $ do
+          "[label](url)" `shouldParseTo`
+            DocParagraph (hyperlink "url" "label")
+
+        it "can be followed by an other paragraph" $ do
+          "[label](url)\n\nfoobar" `shouldParseTo`
+            DocParagraph (hyperlink "url" "label") <> DocParagraph "foobar"
+
+        context "when paragraph contains additional text" $ do
+          it "accepts more text after the link" $ do
+            "[label](url) foo bar baz" `shouldParseTo`
+              DocParagraph (hyperlink "url" "label" <> " foo bar baz")
+
+          it "accepts a newline right after the markdown link" $ do
+            "[label](url)\nfoo bar baz" `shouldParseTo`
+              DocParagraph (hyperlink "url" "label" <> " foo bar baz")
+
+          it "can be followed by an other paragraph" $ do
+            "[label](url)foo\n\nbar" `shouldParseTo`
+              DocParagraph (hyperlink "url" "label" <> "foo") <> DocParagraph "bar"
+
     context "when parsing birdtracks" $ do
       it "parses them as a code block" $ do
         unlines [
@@ -588,7 +682,7 @@ spec = do
 
 
       it "can nest definition lists" $ do
-        "[a] foo\n\n    [b] bar\n\n        [c] baz\n        qux" `shouldParseTo`
+        "[a]: foo\n\n    [b]: bar\n\n        [c]: baz\n        qux" `shouldParseTo`
           DocDefList [ ("a", "foo"
                              <> DocDefList [ ("b", "bar"
                                                    <> DocDefList [("c", "baz\nqux")])
@@ -602,8 +696,25 @@ spec = do
                            ]
           <> DocOrderedList [ DocParagraph "baz" ]
 
+      it "allows arbitrary initial indent of a list" $ do
+        unlines
+          [ "     * foo"
+          , "     * bar"
+          , ""
+          , "         * quux"
+          , ""
+          , "     * baz"
+          ]
+        `shouldParseTo`
+        DocUnorderedList
+          [ DocParagraph "foo"
+          , DocParagraph "bar"
+            <> DocUnorderedList [ DocParagraph "quux" ]
+          , DocParagraph "baz"
+          ]
+
       it "definition lists can come back to top level with a different list" $ do
-        "[foo] foov\n\n    [bar] barv\n\n1. baz" `shouldParseTo`
+        "[foo]: foov\n\n    [bar]: barv\n\n1. baz" `shouldParseTo`
           DocDefList [ ("foo", "foov"
                                <> DocDefList [ ("bar", "barv") ])
                      ]
@@ -751,9 +862,9 @@ spec = do
     context "when parsing definition lists" $ do
       it "parses a simple list" $ do
         unlines [
-            " [foo] one"
-          , " [bar] two"
-          , " [baz] three"
+            " [foo]: one"
+          , " [bar]: two"
+          , " [baz]: three"
           ]
         `shouldParseTo` DocDefList [
             ("foo", "one")
@@ -763,9 +874,9 @@ spec = do
 
       it "ignores empty lines between list items" $ do
         unlines [
-            "[foo] one"
+            "[foo]: one"
           , ""
-          , "[bar] two"
+          , "[bar]: two"
           ]
         `shouldParseTo` DocDefList [
             ("foo", "one")
@@ -773,13 +884,13 @@ spec = do
           ]
 
       it "accepts an empty list item" $ do
-        "[foo]" `shouldParseTo` DocDefList [("foo", DocEmpty)]
+        "[foo]:" `shouldParseTo` DocDefList [("foo", DocEmpty)]
 
       it "accepts multi-line list items" $ do
         unlines [
-            "[foo] point one"
+            "[foo]: point one"
           , "  more one"
-          , "[bar] point two"
+          , "[bar]: point two"
           , "more two"
           ]
         `shouldParseTo` DocDefList [
@@ -788,21 +899,33 @@ spec = do
           ]
 
       it "accepts markup in list items" $ do
-        "[foo] /foo/" `shouldParseTo` DocDefList [("foo", DocEmphasis "foo")]
+        "[foo]: /foo/" `shouldParseTo` DocDefList [("foo", DocEmphasis "foo")]
 
       it "accepts markup for the label" $ do
-        "[/foo/] bar" `shouldParseTo` DocDefList [(DocEmphasis "foo", "bar")]
+        "[/foo/]: bar" `shouldParseTo` DocDefList [(DocEmphasis "foo", "bar")]
 
       it "requires empty lines between list and other paragraphs" $ do
         unlines [
             "foo"
           , ""
-          , "[foo] bar"
+          , "[foo]: bar"
           , ""
           , "baz"
           ]
         `shouldParseTo` DocParagraph "foo" <> DocDefList [("foo", "bar")] <> DocParagraph "baz"
 
+      it "dose not require the colon (deprecated - this will be removed in a future release)" $ do
+        unlines [
+            " [foo] one"
+          , " [bar] two"
+          , " [baz] three"
+          ]
+        `shouldParseTo` DocDefList [
+            ("foo", "one")
+          , ("bar", "two")
+          , ("baz", "three")
+          ]
+
     context "when parsing consecutive paragraphs" $ do
       it "will not capture irrelevant consecutive lists" $ do
         unlines [ "   * bullet"
@@ -815,9 +938,9 @@ spec = do
                 , " "
                 , "   2. different bullet"
                 , "   "
-                , "   [cat] kitten"
+                , "   [cat]: kitten"
                 , "   "
-                , "   [pineapple] fruit"
+                , "   [pineapple]: fruit"
                 ] `shouldParseTo`
           DocUnorderedList [ DocParagraph "bullet"
                            , DocParagraph "different bullet"]
